import asyncio
import time
from typing import *  # type: ignore

import pytest

import imy.async_utils

T = TypeVar("T")


_count_concurrent_running = 0


async def _async_generator(n: int) -> AsyncIterable[int]:
    """
    Like `range`, but asynchronous.
    """
    for ii in range(n):
        yield ii


async def _count_concurrent(value: T) -> tuple[T, int]:
    """
    Returns the input value as is, as well as how often this function was
    running concurrently during the invocation. In order for this to work the
    function imposes a small delay.
    """
    global _count_concurrent_running
    _count_concurrent_running += 1

    try:
        await asyncio.sleep(0.1)
        return value, _count_concurrent_running
    finally:
        _count_concurrent_running -= 1


@pytest.mark.asyncio
async def test_collect_basic() -> None:
    result = await imy.async_utils.collect(_async_generator(5))
    assert result == [0, 1, 2, 3, 4]


@pytest.mark.asyncio
async def test_collect_limit_too_high() -> None:
    result = await imy.async_utils.collect(_async_generator(5), limit=10)
    assert result == [0, 1, 2, 3, 4]


@pytest.mark.asyncio
async def test_collect_limit_too_low() -> None:
    result = await imy.async_utils.collect(_async_generator(5), limit=3)
    assert result == [0, 1, 2]


@pytest.mark.asyncio
async def test_collect_empty() -> None:
    result = await imy.async_utils.collect(_async_generator(0))
    assert result == []


@pytest.mark.asyncio
async def test_amap_simple() -> None:
    results = [
        ii
        async for ii in imy.async_utils.amap(
            _count_concurrent,
            range(5),
            concurrency=100,
        )
    ]

    for ii, (value, concurrent) in enumerate(results):
        assert value == ii


@pytest.mark.asyncio
async def test_amap_concurrency() -> None:
    results = [
        ii
        async for ii in imy.async_utils.amap(
            _count_concurrent,
            range(5),
            concurrency=2,
        )
    ]

    for ii, (value, concurrent) in enumerate(results):
        assert value == ii
        assert concurrent <= 2


# TODO: Test that `amap` returns exceptions ASAP


@pytest.mark.asyncio
async def test_iterator_to_thread_simple() -> None:
    results = [
        val
        async for val in imy.async_utils.iterator_to_thread(
            range(5),
            batch_size=5,
        )
    ]

    for value_should, value_is in enumerate(results):
        assert value_should == value_is


# TODO: More tests for `iterator_to_thread`




# fake module to reside in my lib folder and import the actual implementation

from pathlib import Path

here = Path(__file__).absolute().parent
name = here.stem


import sys
sys.path.insert(0, str(here))
del sys.modules[name]
module = __import__(name)
del sys.path[0]

del Path
del here
del name
del sys
globals().update(module.__dict__)



"""
Utilities for getting metadata about your own package.
"""

import importlib.metadata
import inspect
from pathlib import Path
from typing import *  # type: ignore


def _get_root_directory(caller: inspect.FrameInfo) -> Path:
    """
    Return the root directory of the calling module, accounting for submodules.

    For example, if the calling module is `foo.bar.baz`, and `foo` is located
    at `/home/user/foo`, this function will return `/home/user/foo`.
    """
    caller_path = Path(caller.filename).absolute()

    # `__init__` files aren't their own submodules, so treat them as their
    # parent
    if caller_path.name == "__init__.py":
        caller_path = caller_path.parent

    # How deep is this submodule?
    depth: int = caller.frame.f_globals["__name__"].count(".")

    for _ in range(depth):
        caller_path = caller_path.parent

    return caller_path


def _find_pyproject_toml(module_directory: Path) -> Path:
    """
    Attempts to locate a project's `pyproject.toml` file. It will look in common
    locations and return the path if found, or raise a `FileNotFoundError` if
    not.
    """

    def _iter_toml_paths() -> Iterable[Path]:
        # In a typicical project the toml is right next to the module
        yield module_directory.parent / "pyproject.toml"

        # However, some projects have a `src` directory
        if module_directory.parent.name == "src":
            yield module_directory.parent.parent / "pyproject.toml"

    # Try to find the toml
    for toml_path in _iter_toml_paths():
        if toml_path.exists():
            return toml_path

    # Nothing was found
    raise FileNotFoundError("Could not find `pyproject.toml`")


def get_package_version(own_package_pypi_name: str) -> str:
    """
    Determine the version of **the calling package**. `own_package_pypi_name` is
    the name of the package, as you would type to install it from pypi.

    You can **not use this function to look up any packages other than your
    own**.
    """

    # Try to just ask python
    try:
        return importlib.metadata.version(own_package_pypi_name)
    except importlib.metadata.PackageNotFoundError:
        pass

    # While the approach above is clean, it fails during development. In that
    # case, read the version from the `pyproject.toml` file.
    import tomllib

    caller_frame = inspect.stack()[1]
    toml_path = _find_pyproject_toml(_get_root_directory(caller_frame))

    try:
        with open(toml_path, "rb") as f:
            toml_contents = tomllib.load(f)

    except FileNotFoundError:
        raise RuntimeError(f"Cannot find `pyproject.toml` at `{toml_path}`") from None

    except tomllib.TOMLDecodeError as e:
        raise RuntimeError(f"`{toml_path}` is invalid TOML: {e}") from None

    # The version can be stored in different places, depending on the tooling
    try:
        return toml_contents["project"]["version"]
    except KeyError:
        pass

    try:
        return toml_contents["tool"]["poetry"]["version"]
    except KeyError:
        pass

    raise RuntimeError(
        f"`{toml_path}` does not contain a `tool.poetry.version` field"
    ) from None



import asyncio
import threading
from typing import *  # type: ignore

T = TypeVar("T")
IN = TypeVar("IN")
OUT = TypeVar("OUT")


async def collect(
    values: AsyncIterable[IN],
    *,
    limit: int | None = None,
) -> list[IN]:
    """
    Collects the values from an asynchronous iterable into a list.

    If `limit` is set, the result will contain at most the first `limit` values.

    ## Parameters

    `values`: The asynchronous iterable to collect values from.

    `limit`: The maximum number of values to collect. If `None`, all values will
        be collected.
    """
    result: list[IN] = []

    async for value in values:
        result.append(value)

        if len(result) == limit:
            break

    return result


async def amap(
    func: Callable[[IN], Awaitable[OUT]],
    values: Iterable[IN],
    *,
    concurrency: int = 10,
) -> AsyncIterable[OUT]:
    """
    Asynchronously maps `func` over `values`, with at most `concurrency` items
    being processed simultaneously.

    ## Parameters

    `func`: The function to apply to each value.

    `values`: The values to apply the function to.

    `concurrency`: The maximum number of items to process simultaneously.
    """
    # TODO: Support async iterables for the input

    # Processes a single value. Returns both the index and the result.
    async def process_single(index: int, value: IN) -> tuple[int, OUT]:
        result = await func(value)
        return index, result

    # Yield values as they become available - however, make sure to raise asap
    # if any task fails
    to_do: list[tuple[int, IN]] = list(enumerate(values))
    in_flight: set[asyncio.Task[tuple[int, OUT]]] = set()

    n_already_yielded: int = 0
    not_yet_yielded: dict[int, OUT] = {}

    while True:
        # Spawn tasks until the limit is reached
        while len(in_flight) < concurrency:
            try:
                index, value = to_do.pop(0)
            except IndexError:
                break

            in_flight.add(
                asyncio.create_task(
                    process_single(
                        index,
                        value,
                    )
                )
            )

        # Done?
        if not in_flight and not to_do:
            break

        # Wait for any task to complete
        done, in_flight = await asyncio.wait(
            in_flight,
            return_when=asyncio.FIRST_COMPLETED,
        )

        # Add any completed tasks to the results
        for task in done:
            # Pass-through exceptions
            try:
                index, result = task.result()
            except Exception:
                for task in in_flight:
                    task.cancel()

                raise

            # Store the result
            assert index >= n_already_yielded, (
                index,
                n_already_yielded,
                not_yet_yielded,
            )
            not_yet_yielded[index] = result

        # Yield any results that are now available
        while True:
            try:
                value = not_yet_yielded.pop(n_already_yielded)
            except KeyError:
                break
            else:
                n_already_yielded += 1
                yield value


async def iterator_to_thread(
    sync_iterable: Iterable[T],
    *,
    batch_size: int,
) -> AsyncIterable[T]:
    """
    Given a potentially slow, synchronous iterator, returns an equivalent
    asynchronous iterator.

    The synchronous iterator will be run in a separate thread, and the results
    retrieved asynchronously. To decrease overhead, values from the synchronous
    iterator are fetched in batches. This means that values won't be yielded
    until `batch_size` values have been collected, or the iterator is exhausted.
    If this behavior isn't desirable in your application, set `batch_size` to
    `1`.

    Exceptions from the synchronous iterator will be propagated.
    """
    assert batch_size > 0, "Batch size must be positive"

    mainloop = asyncio.get_event_loop()
    results: asyncio.Queue[
        tuple[Literal["values"], list[T]]
        | tuple[Literal["error"], Exception]
        | tuple[Literal["done"]]
    ] = asyncio.Queue()
    stop_requested = False

    # The worker function continuously fills the queue with results. It is
    # limited by the buffer size.
    def worker():
        sync_iterator = iter(sync_iterable)
        batch: list[T] = []

        while True:
            # Try to get some values from the iterator
            try:
                batch.append(next(sync_iterator))

            # Done?
            except StopIteration:
                mainloop.call_soon_threadsafe(results.put_nowait, ("values", batch))
                mainloop.call_soon_threadsafe(results.put_nowait, ("done",))
                return

            # Handle errors
            except Exception as e:
                mainloop.call_soon_threadsafe(results.put_nowait, ("values", batch))
                mainloop.call_soon_threadsafe(results.put_nowait, ("error", e))
                return

            # If the batch is full, pass it on
            if len(batch) >= batch_size:
                mainloop.call_soon_threadsafe(results.put_nowait, ("values", batch))
                batch = []

            # If the stop was requested, pass on the remaining values
            if stop_requested:
                mainloop.call_soon_threadsafe(results.put_nowait, ("values", batch))
                mainloop.call_soon_threadsafe(results.put_nowait, ("done",))
                return

    worker_thread = threading.Thread(target=worker, daemon=True)

    # The remainder of the function has to be wrapped to ensure the thread is
    # stopped if anything goes wrong.
    try:
        worker_thread.start()

        while True:
            # Wait for the next result
            result = await results.get()

            # Yield the values
            if result[0] == "values":
                for value in result[1]:
                    yield value

            # Handle errors
            elif result[0] == "error":
                raise result[1]

            # Done
            elif result[0] == "done":
                return

            # Invalid
            else:
                assert False, f"Received invalid result `{result!r}`"

    finally:
        stop_requested = True
